# Functions for reading ascents from theCrag's CSV logbook exports.

# Copyright Contributors to the Climbing Ratings project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#' Parses the number of pitches logged, from a logbook's Comment field.
#'
#' @param comment character value of the Comment field from the logbook export.
#' This is assumed to be a singleton.
#' @return an integer representing the number of pitches with info logged, or
#' NA_integer_ if no pitch info was logged.  Note this may not reflect the
#' actual number of pitches on the route.
#' @keywords internal
.GetPitchCount <- function(comment) {
  # If there is pitch info for an ascent, theCrag stuffs it into the comment
  # field.  This isn't very reliable: if it's a multipitch route but the
  # climber didn't fill out pitch data for the ascent, it will be missing.
  # As at April 2020, it also doesn't include the pitch-tick, making it useless
  # for pitch-level estimation purposes.
  matches <- comment %>%
    stringr::str_split(stringr::fixed("\n")) %>%
    unlist() %>%
    stringr::str_match("(\\d+):(?:\\[[^]]+\\])?")

  n <- nrow(matches)

  # Find the index of the last comment line.
  i <- utils::tail(c(0L, which(is.na(matches[, 1]))), n = 1)

  if (0L < n - i) {
    p <- as.integer(matches[(i + 1):n, 2])
    if (!is.unsorted(p, strictly = TRUE)) {
      return(1L + p[[length(p)]] - p[[1]])
    }
  }

  NA_integer_
}

#' Reads a CSV export of a climber's logbook from theCrag.
#'
#' @param df data frame with the same columns as CSV logbooks from theCrag.
#' @param climber character username to populate the "climber" column in the
#' returned data frame.
#' @return a "raw ascents" data frame (see [ReadLogbooks()]).
#' @keywords internal
.ParseLogbook <- function(df, climber) {
  df %>%
    dplyr::arrange(.data$Log.Date) %>%
    dplyr::transmute(
      ascentId = .data$Ascent.ID,
      route = .data$Route.ID,
      climber = climber,
      tick = NormalizeAscentType(.data$Ascent.Type),
      grade = suppressWarnings(as.integer(.data$Route.Grade)),
      timestamp = .data$Ascent.Date %>%
        as.POSIXct(format = "%FT%H:%M:%SZ", optional = TRUE, tz = "UTC") %>%
        as.integer(),
      style = 1L,
      pitches = purrr::map_int(.data$Comment, .GetPitchCount)
    ) %>%
    tidyr::drop_na(-.data$pitches)
}

#' Reads a CSV export of a climber's logbook from theCrag.
#'
#' @param filename path to file, assumed to be in the format generated by
#' theCrag, i.e. "climber-logbook-YYYY-MM-DD.csv".
#' @return a "raw ascents" data frame (see [ReadLogbooks()]).
ReadLogbook <- function(filename) {
  m <- stringr::str_match(
    basename(filename),
    "([^-]+)-logbook-\\d{4}-\\d{2}-\\d{2}"
  )
  utils::read.csv(
    filename,
    # R doesn't support 64-bit integers, so force IDs to be interpreted as characters.
    colClasses = c(Ascent.ID = "character", Route.ID = "character"),
    stringsAsFactors = FALSE
  ) %>%
    .ParseLogbook(climber = stringr::str_to_lower(m[[1, 2]]))
}

#' Reads all logbooks in a directory.  Logbook filenames are assumed to be
#' in the format generated by theCrag, i.e. "climber-logbook-YYYY-MM-DD.csv".
#'
#' @param dir the directory containing logbook files to read.
#' @return a "raw ascents" data frame.  This contains one row per ascent,
#' and has the columns:
#'
#' * ascentId: character ID of the ascent
#' * route: character ID of the route
#' * climber: character normalized climber (derived from the filename)
#' * tick: character normalized tick type
#' * grade: integer grade (e.g. Ewbank's grade)
#' * style: integer always 0 (column present for consistency with JSON)
#' * timestamp: integer seconds since UNIX epoch at date of ascent
#' * pitches: integer or NA number of pitches logged
ReadLogbooks <- function(dir) {
  logbooks <- Sys.glob(file.path(dir, "*-logbook-*.csv"))
  dfs <- purrr::map(logbooks, ReadLogbook)
  dplyr::bind_rows(dfs)
}
